/*
Copyright (C) 2009  Ruan Strydom <ruan@jcell.co.za>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "cpppropertycreator.h"
#include "propertynode.h"
#include "propertymodel.h"

#include <language/duchain/duchainlock.h>
#include <language/duchain/duchain.h>
#include <language/duchain/classdeclaration.h>
#include <language/duchain/declaration.h>
#include <language/duchain/duchainutils.h>
#include <language/duchain/classfunctiondeclaration.h>
#include <language/duchain/types/functiontype.h>
#include <interfaces/idocument.h>
#include <interfaces/icore.h>
#include <interfaces/idocumentcontroller.h>

#include <KTextEditor/View>
#include <KTextEditor/Document>
#include <interfaces/isourceformattercontroller.h>
#include <interfaces/isourceformatter.h>
#include <interfaces/ilanguagecontroller.h>
#include <interfaces/ilanguage.h>
#include <language/interfaces/ilanguagesupport.h>
#include <language/duchain/types/typeutils.h>
#include <KLocalizedString>

#define ERROR_MESSAGE ({KMessageBox::error(NULL, i18n(QString("An error occured while parsing the class. Please log a report. File: %1 Line: %2").arg(__FILE__).arg(__LINE__).toAscii().constData()));});
#include <KDE/KMessageBox>

CppPropertyCreator::CodeRange* CppPropertyCreator::getRange ( const QString& name )
{
    //There is only a small number so a loop should be fine
    for ( int i = 0; i < m_ranges.size(); i++ )
    {
        if ( m_ranges.at ( i ).name == name )
            return & ( m_ranges[i] );
    }
    return NULL;
}


int CppPropertyCreator::getRangeIndex ( const QString& name )
{
    //There is only a small number so a loop should be fine
    for ( int i = 0; i < m_ranges.size(); i++ )
    {
        if ( m_ranges.at ( i ).name == name )
            return i;
    }
    return -1;
}


void CppPropertyCreator::create ( PropertyModel* m )
{
    DUChainReadLocker lock ( DUChain::lock() );

    ClassDeclaration * cd = m->klass();

    IDocument* doc = ICore::self()->documentController()->documentForUrl ( cd->topContext()->url().toUrl() );
    if ( !doc )
    {
        ERROR_MESSAGE
        return;
    }

    KTextEditor::Document* textDoc = doc->textDocument();
    if ( !textDoc )
    {
        ERROR_MESSAGE
        return;
    }

    KTextEditor::View* view = textDoc->activeView();
    if ( !view )
    {
        ERROR_MESSAGE
        return;
    }

    //Turn on clumsy code

    QString text = textDoc->text ( m->klass()->internalContext()->range().textRange() );

    int lineCount = m->klass()->internalContext()->range().start.line;
    int dummy = 0, *end = &dummy;

    QRegExp publicExp ( "^(\\s*)public\\s*:.*" );
    QRegExp protectedExp ( "^(\\s*)protected\\s*:.*" );
    QRegExp privateExp ( "^(\\s*)private\\s*:.*" );
    QRegExp indentExp ( "^(\\s+)\\w+.*" );
    QStringList lines = text.split ( '\n' );

    QString indent = "";
    QString accessIndent = "";

    m_ranges.clear();
    m_ranges.append ( CodeRange ( "head", lineCount, lineCount ) );
    end = & ( getRange ( "head" )->lineEnd );

    foreach ( QString line, lines )
    {
        //Skip inline classes
        DUContext *duc = m->klass()->internalContext()->findContextAt ( SimpleCursor ( lineCount, 0 ) );
        if ( duc && duc != m->klass()->internalContext() )
        {
            lineCount++;
            continue;
        }

        if ( publicExp.exactMatch ( line ) )
        {
            if ( !getRange ( "public" ) )
            {
                m_ranges.append ( CodeRange ( "public" ) );
            }
            getRange ( "public" )->lineStart = lineCount;
            *end = lineCount;
            end = & ( getRange ( "public" )->lineEnd );
            accessIndent = publicExp.cap ( 1 );
        }
        else if ( protectedExp.exactMatch ( line ) )
        {
            if ( !getRange ( "protected" ) )
            {
                m_ranges.append ( CodeRange ( "protected" ) );
            }
            getRange ( "protected" )->lineStart = lineCount;
            *end = lineCount;
            end = & ( getRange ( "protected" )->lineEnd );
            accessIndent = protectedExp.cap ( 1 );
        }
        else if ( privateExp.exactMatch ( line ) )
        {
            if ( !getRange ( "private" ) )
            {
                m_ranges.append ( CodeRange ( "private" ) );
            }
            getRange ( "private" )->lineStart = lineCount;
            *end = lineCount;
            end = & ( getRange ( "private" )->lineEnd );
            accessIndent = privateExp.cap ( 1 );
        }
        else if ( indentExp.exactMatch ( line ) )
        {
            indent = indentExp.cap ( 1 );
        }

        lineCount++;
    }
    lineCount--;
    *end = m->klass()->internalContext()->range().end.line;

    foreach ( PropertyNode n, m->nodes() )
    {
        if ( n.createReader() && n.canCreateReader() )
        {
            QStringList itext ( "" );
            QString text = QString ( "%1 %2() const { return %3; }" ).arg ( readerType ( n.type(), n.resolvedType(), n.whichType() ), n.readerName(), n.name() );

            switch ( m->modifier() )
            {
            case PropertyModel::VIRTUAL:
                text = QString ( "virtual " ) + text;
                break;
            case PropertyModel::PURE_VIRTUAL:
                text = QString ( "virtual %1 %2() const = 0;" ).arg ( readerType ( n.type(), n.resolvedType(), n.whichType() ), n.readerName() );
                break;
            case PropertyModel::INLINE:
                text = QString ( "inline " ) + text;
                break;
            }

            CodeRange *range = getRange ( n.readerAccess() );
            if ( !range )
            {
                m_ranges.append ( CodeRange ( n.readerAccess(), lineCount, lineCount ) );
                range = getRange ( n.readerAccess() );
                addText ( doc, range, QStringList ( QString ( "%1:" ).arg ( n.readerAccess() ) ), accessIndent, &lineCount );
            }

            if ( m->createDocumentation() )
            {
                itext.append ( QString ( "/** Reader for %1." ).arg ( n.name() ) );
                itext.append ( QString ( " *  @see %1" ).arg ( n.name() ) );
                itext.append ( QString ( " *  @return %1" ).arg ( n.name() ) );
                itext.append ( " */" );
            }
            itext.append ( text );
            addText ( doc, range, itext, indent, &lineCount );
        }

        if ( n.createWriter() && n.canCreateWriter() )
        {
            QStringList itext ( "" );
            QString text = QString ( "void %1( %2 %4 ) { %3 = %4; }" ).arg ( n.writerName(), writerType ( n.type(), n.resolvedType(), n.whichType() ), n.name(), propertyAlias ( n.name() ) );

            //We cant asign arrays, it needs a memcpy or something
            if ( n.whichType() == AbstractType::TypeArray )
            {
                static QRegExp e ( "(.*)(\\[(\\d*)\\])" );
                e.exactMatch ( n.type() );
                text = QString ( "void %1( const %2 %3%4 ) { /*TODO*/ for(int i; i<%5; i++) %6[i] = %3[i]; }" ).arg ( n.writerName(), e.cap ( 1 ), propertyAlias ( n.name() ), e.cap ( 2 ), e.cap ( 3 ), n.name() );
            }

            switch ( m->modifier() )
            {
            case PropertyModel::VIRTUAL:
                text = QString ( "virtual " ) + text;
                break;
            case PropertyModel::PURE_VIRTUAL:
                text = QString ( "virtual void %1( %2 %3 ) = 0;" ).arg ( n.writerName(), writerType ( n.type(), n.resolvedType(), n.whichType() ), propertyAlias ( n.name() ) );
                //We cant asign arrays, it needs a memcpy or something
                if ( n.whichType() == AbstractType::TypeArray )
                {
                    static QRegExp e ( "(.*)(\\[\\d*\\])" );
                    e.exactMatch ( n.type() );
                    text = QString ( "void %1( const %2 %3%4 ) = 0;" ).arg ( n.writerName(), e.cap ( 1 ), propertyAlias ( n.name() ), e.cap ( 2 ) );
                }
                break;
            case PropertyModel::INLINE:
                text = QString ( "inline " ) + text;
                break;
            }

            CodeRange *range = getRange ( n.writerAccess() );
            if ( !range )
            {
                m_ranges.append ( CodeRange ( n.writerAccess(), lineCount, lineCount ) );
                range = getRange ( n.writerAccess() );
                addText ( doc, range, QStringList ( QString ( "%1:" ).arg ( n.writerAccess() ) ), accessIndent, &lineCount );
            }

            if ( m->createDocumentation() )
            {
                itext.append ( QString ( "/** Writer for %1." ).arg ( n.name() ) );
                itext.append ( QString ( " *  @see %1" ).arg ( n.name() ) );
                itext.append ( QString ( " *  @param %1 Value for %2." ).arg ( propertyAlias ( n.name() ), n.name() ) );
                itext.append ( " */" );
            }

            itext.append ( text );
            addText ( doc, range, itext, indent, &lineCount );
        }

        if ( m->hasQObject() && m->createQtProperties() && ( ( n.createReader() && n.canCreateReader() ) || ( n.createWriter() && n.canCreateWriter() ) ) )
        {
            QStringList itext;
            QString text = QString ( "Q_PROPERTY(%1 %2%3%4)" ).arg ( n.type(),
                           propertyName ( n.name() ),
                           n.createReader() && n.canCreateReader() ? QString ( " READ %1" ).arg ( readerName ( n.name() ) ) : "",
                           n.createWriter() && n.canCreateWriter() ? QString ( " WRITE %1" ).arg ( writerName ( n.name() ) ) : "" );
            CodeRange *range = getRange ( "head" );
            itext.append ( text );
            addText ( doc, range, itext, indent, &lineCount );
        }
    }

    //TODO  while(true) { Turn off clumsy code!! }
}


void CppPropertyCreator::addText ( IDocument *doc, CodeRange *range, const QStringList &text, const QString &indent, int *lineCount )
{
    if ( !doc || !range )
    {
        ERROR_MESSAGE
        return;
    }
    KTextEditor::Document* textDoc = doc->textDocument();
    if ( !textDoc )
    {
        ERROR_MESSAGE
        return;
    }

    QStringList itext;

    if ( ICore::self() )
    {
        QString t = ICore::self()->sourceFormatterController()->formatterForUrl ( doc->url() )->formatSource ( text.join ( "\n" ) ,doc->mimeType() );
        itext = t.split ( "\n" );
    }
    while ( itext.last().size() == 0 || itext.last() == "\n" )
        itext.removeLast();

    for ( int i= 0; i < itext.size(); i++ )
    {
        textDoc->insertLine ( range->lineEnd, indent + itext[i] );
        range->lineEnd+=1;
        if ( lineCount )
            *lineCount = *lineCount + 1;
    }
    for ( int i = getRangeIndex ( range->name ) + 1; i < m_ranges.size(); i++ )
    {
        m_ranges[i].lineStart += itext.size();
        m_ranges[i].lineEnd += itext.size();
    }
}

bool CppPropertyCreator::parse ( PropertyModel* m, QList<Declaration*> selection )
{
    DUChainReadLocker lock ( DUChain::lock() );
    QList<PropertyNode> nodes;

    //Damn - just making sure we get a class.
    ClassDeclaration *klass = NULL;
    foreach ( Declaration* decl,  selection )
    {
        if ( decl && decl->context() && decl->context()->type() == DUContext::Class )
        {
            klass = dynamic_cast<ClassDeclaration *> ( decl->context()->owner() );
            if ( klass )
                break;
        }
        else if ( decl && decl->internalContext() && decl->internalContext()->type() == DUContext::Class )
        {
            klass = dynamic_cast<ClassDeclaration *> ( decl->internalContext()->owner() );
            if ( klass )
                break;
        }
        klass = dynamic_cast<ClassDeclaration *> ( decl );
        if ( klass )
            break;
    }

    //We did not?
    if ( !klass )
    {
        ERROR_MESSAGE
        return false;
    }

    m->setKlass ( klass );

    foreach ( Declaration* decl, klass->internalContext()->localDeclarations() )
    {
        ClassMemberDeclaration* memDecl = dynamic_cast<ClassMemberDeclaration*> ( decl );
        if ( !memDecl )
            continue;

        if ( ! ( !memDecl->isFunctionDeclaration() && memDecl->kind() == Declaration::Instance ) )
            continue;

        //TODO This depends on staticMetaObject being at the begining of the loop
        if ( memDecl->identifier().toString() == "staticMetaObject" )
        {
            m->setHasQObject ( true );
            continue;
        }

        PropertyNode n;
        n.setName ( memDecl->identifier().toString() );
        n.setType ( memDecl->abstractType()->toString() );
        AbstractType::Ptr aptr = TypeUtils::unAliasedType ( memDecl->abstractType() );
        n.setResolvedType ( aptr->toString() );
        n.setWhichType ( aptr->whichType() );
        n.setReaderName ( readerName ( n.name() ) );
        n.setWriterName ( writerName ( n.name() ) );

        /*(The users selection) NOTE I dont know if comparing the pointers is valid?
            But the duchain was locked before calling this routine so its should be.
        */
        foreach ( Declaration* decl,  selection )
        {
            if ( memDecl == decl )
            {
                n.setCreateReader ( true );
                n.setCreateWriter ( true );
            }
        }

        //ie: const int& TODO get modifier from duchain rather
        if ( (n.whichType() == AbstractType::TypeReference || n.whichType() == AbstractType::TypeArray) &&
            ( n.type().contains ( "const" ) || n.resolvedType().contains ( "const" ) ) )
        {
            n.setCanCreateWriter ( false );
            n.setCreateWriter ( false );
        }

        foreach ( Declaration * d, klass->internalContext()->localDeclarations() )
        {
            ClassFunctionDeclaration * dm =  dynamic_cast<ClassFunctionDeclaration*> ( d );
            if ( !dm )
                continue;

            FunctionType::Ptr ft = dm->abstractType().cast<FunctionType>();
            if ( !ft )
                continue;
            if ( !ft->returnType() )
                continue;

            QString access = "default";
            switch ( dm->accessPolicy() )
            {
            case Declaration::Public:
                access = "public";
                break;
            case Declaration::Protected:
                access = "protected";
                break;
            case Declaration::Private:
                access = "private";
                break;
            }

            if ( dm->identifier().toString() == n.readerName() &&
                    ft->arguments().size() == 0 &&
                    ft->modifiers() & AbstractType::ConstModifier )
            {
                n.setCanCreateReader ( false );
                n.setCreateReader ( true );
                n.setReaderAccess ( access );
                continue;
            }

            if ( dm->identifier().toString() == n.writerName() &&
                    ft->returnType()->toString() == "void" &&
                    ft->arguments().size() == 1 &&
                    ft->arguments() [0]->toString() == writerType ( n.type(), n.resolvedType(), n.whichType() ) )
            {
                n.setCanCreateWriter ( false );
                n.setCreateWriter ( true );
                n.setWriterAccess ( access );
                continue;
            }
        }
        nodes.append ( n );
    }

    m->setNodes ( nodes );

    return true;
}


QString CppPropertyCreator::propertyAlias ( const QString& property )
{
    if ( property.isNull() )
    {
        ERROR_MESSAGE
        return "null";
    }

    QString p = propertyName ( property );
    QString r = "";
    r += p[0];
    bool kase = false;
    for ( int i = 1; i < p.size(); i++ )
    {
        if ( p[i].isUpper() && kase == false )
        {
            r += p[i].toLower();
            kase = true;
        }
        if ( p[i].isLower() && kase == true )
            kase = false;
        if ( p[i] == '_' && i < p.size() - 1 )
            r += p[i+1];
    }

    if ( r == property )
        r = QString ( "_%1" ).arg ( r );
    return r;
}


QString CppPropertyCreator::propertyName ( const QString& property )
{
    if ( property.isNull() )
    {
        ERROR_MESSAGE
        return "null";
    }

    QString n = QString ( "%1" ).arg ( property );
    if ( property.size() > 1 && property[0] == 'm' && property[1].isUpper() )
        n = n.replace ( QRegExp ( "(^m_)|(^m)|(^_+)" ), "" );
    else
        n = n.replace ( QRegExp ( "(^m_)|(^_+)" ), "" );
    n[0] = n[0].toLower();
    return n;
}

QString CppPropertyCreator::readerName ( const QString& property )
{
    if ( property.isNull() )
    {
        ERROR_MESSAGE
        return "null";
    }

    QString n = propertyName ( property );
    if ( n == property )
    {
        n[0] = n[0].toUpper();
        n = QString ( "get%1" ).arg ( n );
    }
    return n;
}

QString CppPropertyCreator::writerName ( const QString& property )
{
    if ( property.isNull() )
    {
        ERROR_MESSAGE
        return "null";
    }

    QString n = propertyName ( property );
    n[0] = n[0].toUpper();
    return QString ( "set%1" ).arg ( n );
}

QString CppPropertyCreator::readerType ( const QString & abstractType, const QString & resolvedType, AbstractType::WhichType type )
{
    if ( abstractType.isNull() ||  resolvedType.isNull())
    {
        ERROR_MESSAGE
        return "null";
    }

    //TODO This could and probably should come from the DUChain 'modifier'
    bool hasConst = abstractType.contains ( "const" ) || resolvedType.contains ( "const" );

    //Primitives
    if ( type == AbstractType::TypeIntegral || type == AbstractType::TypeEnumeration )
        return QString ( abstractType ).replace ( "const", "" );

    //Array's see writerType()
    if ( type == AbstractType::TypeArray )
    {
        if ( hasConst )
            return QString ( "%1*" ).arg ( QString ( abstractType ).replace ( QRegExp ( "(\\[\\d*\\])" ),"" ) );
        else
            return QString ( "const %1*" ).arg ( QString ( abstractType ).replace ( QRegExp ( "(\\[\\d*\\])" ),"" ) );
    }

    //References and pointer
    if ( type == AbstractType::TypePointer || type == AbstractType::TypeReference )
    {
        if ( hasConst )
            return  QString ( abstractType );
        else
            return  QString ( "const %1" ).arg ( abstractType );
    }

    //Whats left
    if ( hasConst )
        return QString ( "%1&" ).arg ( abstractType );
    return QString ( "const %1&" ).arg ( abstractType );
}

QString CppPropertyCreator::writerType ( const QString & abstractType, const QString & resolvedType, AbstractType::WhichType type )
{
    if ( abstractType.isNull() ||  resolvedType.isNull())
    {
        ERROR_MESSAGE
        return "null";
    }

    //TODO This could and probably should come from the DUChain 'modifier'
    bool hasConst = abstractType.contains ( "const" ) || resolvedType.contains ( "const" );

    //Arrays need a special implementation, use this to compare parameters
    if ( type == AbstractType::TypeArray )
        return QString("const %1").arg(abstractType);
    
    //Primitives
    if ( type == AbstractType::TypeIntegral || type == AbstractType::TypeEnumeration )
        return QString ( abstractType ).replace ( "const", "" );

    //References and pointer
    if ( type == AbstractType::TypePointer || type == AbstractType::TypeReference )
        return  QString ( "%1" ).arg ( abstractType );

    //Whats left
    if ( hasConst )
        return QString ( "%1&" ).arg ( abstractType );
    return QString ( "const %1&" ).arg ( abstractType );
}

// kate: indent-mode cstyle; space-indent on; indent-width 4; replace-tabs on;  replace-tabs on;  replace-tabs on;
