/*
Copyright (C) 2009  Ruan Strydom <ruan@jcell.co.za>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "readerwriterplugin.h"

#include <interfaces/icore.h>
#include <interfaces/contextmenuextension.h>
#include <interfaces/context.h>
#include <interfaces/idocument.h>
#include <interfaces/iuicontroller.h>
#include <interfaces/idocumentcontroller.h>

#include <language/duchain/declaration.h>
#include <language/duchain/duchainutils.h>
#include <language/duchain/duchain.h>
#include <language/duchain/duchainlock.h>
#include <language/duchain/classdeclaration.h>
#include <language/duchain/classmemberdeclaration.h>
#include <shell/sourceformattercontroller.h>
#include <interfaces/isourceformatter.h>

#include <KPluginFactory>
#include <KPluginLoader>
#include <KLocalizedString>
#include <KAboutData>
#include <KTextEditor/Document>
#include <KDebug>
#include <KMessageBox>
#include <KActionCollection>
#include <KAction>

#include <QtCore/QPair>
#include <QtCore/QList>

#include "propertiesdialog.h"
#include "propertymodel.h"
#include <ktexteditor/smartcursor.h>
#include "cpppropertycreator.h"


#define ERROR_MESSAGE ({KMessageBox::error(NULL, i18n(QString("An error occured while parsing the class. Please log a report. File: %1 Line: %2").arg(__FILE__).arg(__LINE__).toAscii().constData()));});

K_PLUGIN_FACTORY (KdevPlugFactory, registerPlugin<ReaderWriterPlugin>();)
K_EXPORT_PLUGIN (KdevPlugFactory (KAboutData ("ReaderWriter","ReaderWriter", ki18n ("Generate Reader Writer Properties"), "0.1", ki18n ("Generate Reader Writer Properties"), KAboutData::License_GPL)))

ReaderWriterPlugin::ReaderWriterPlugin (QObject *parent, const QVariantList &)
        :IPlugin (KdevPlugFactory::componentData(), parent)
{
    setXMLFile ("readerwriter.rc");

    m_action = actionCollection()->addAction ("generate_readerwriter");
    m_action->setIcon (KIcon ("kdevelop"));
    m_action->setText (i18n ("Generate properties.."));
    m_action->setShortcut (KShortcut ("Ctrl+Alt+p"));
    connect (m_action, SIGNAL (triggered (bool)), this, SLOT (slotCreateProperties()));
    m_action->setToolTip (i18n ("Generate Reader Writer Properties"));
    m_action->setWhatsThis (i18n ("<b>ReaderWriter</b><p> Generate Reader Writer Properties"));
    m_action->setEnabled (false);
}

ReaderWriterPlugin::~ReaderWriterPlugin()
{
}

void ReaderWriterPlugin::slotCreateProperties()
{
    DUChainReadLocker lock (DUChain::lock());

    QList<Declaration*> decls = cursorItemDeclarations();
    if (decls.size() < 1)
    {
        KMessageBox::information (0, i18n ("No valid declaration in selection.\nIt is possible that the parser is still bussy."));
        return;
    }
    
    createProperties (decls);
}

void ReaderWriterPlugin::createProperties (QList<Declaration*> decls)
{
    DUChainReadLocker lock (DUChain::lock());
    static PropertiesDialog *d = (
    {
        PropertiesDialog *d = new PropertiesDialog ( (QWidget *) ICore::self()->uiController()->activeMainWindow());
        connect (d, SIGNAL (okClicked()), this, SLOT (slotWriteProperties()));
        d;
    });
    static PropertyModel *m = new PropertyModel;
    static CppPropertyCreator pc;

    if (d->isVisible())
        return;
    pc.parse (m, decls);
    d->setCaption (i18n ("Class: ") + m->klass()->identifier().toString());
    d->setModal (false);
    d->show (m);
}

void ReaderWriterPlugin::slotWriteProperties()
{
    static CppPropertyCreator pc;
    PropertiesDialog *d = dynamic_cast<PropertiesDialog *> (this->sender());
    if (!d)
    {
        ERROR_MESSAGE
        kDebug() << "sender() not instance of PropertiesDialog";
        return;
    }
    pc.create (d->model());
}

ContextMenuExtension ReaderWriterPlugin::contextMenuExtension (Context* context)
{
    Q_UNUSED (context);

    DUChainReadLocker lock (DUChain::lock());
    
    m_action->setEnabled (false);
    ContextMenuExtension ext;

    QList<Declaration *> decls = cursorItemDeclarations();

    if (decls.size() > 0) {
        ext.addAction (ContextMenuExtension::EditGroup, m_action);
        m_action->setEnabled (true);
    }

    return ext;
}

QList<Declaration*> ReaderWriterPlugin::cursorItemDeclarations()
{
    QList<Declaration*> list;
    IDocument* doc = ICore::self()->documentController()->activeDocument();
    if (!doc)
        return list;

    KTextEditor::Document* textDoc = doc->textDocument();
    if (!textDoc)
        return list;

    KTextEditor::View* view = textDoc->activeView();
    if (!view)
        return list;

    if(view->selection()) {
        KTextEditor::Range r = view->selectionRange();
        TopDUContext * tc = DUChainUtils::standardContextForUrl(doc->url());
        for(int i = r.start().line(); i <= r.end().line(); i++) {
            Declaration *d = DUChainUtils::declarationInLine(SimpleCursor(i,0),tc);
            if (d && d->kind() == Declaration::Instance && !d->isFunctionDeclaration() &&
                d->context() && d->context()->type() == DUContext::Class)
            {
                list.append(d);
            }
        }
    } else {
        SimpleCursor c (view->cursorPosition());
        Declaration* d = DUChainUtils::itemUnderCursor (doc->url(), c);
        if (d && d->kind() == Declaration::Instance && !d->isFunctionDeclaration() &&
            d->context() && d->context()->type() == DUContext::Class)
            list.append(d);
        else if (d && d->internalContext() && d->internalContext()->type() == DUContext::Class)
            list.append(d);
    }
    return list;
}

#include "readerwriterplugin.moc"

// kate: indent-mode cstyle; space-indent on; indent-width 4; replace-tabs on; 
